/*global console */
var util = require( 'util' ),
	f = util.format,
	chalk = require( 'chalk' ),
	stevedore = require( 'stevedore' ),
	ansiRegex = require( 'ansi-regex' ),
	frames = require( './frames' ),
	summariseChanges = require( './summariseChanges' ),
	tryInstallPlugin = require( './tryInstallPlugin' ),
	messages = '',
	logger,
	messages,
	warnings,
	alreadyWarned = {},
	errorHandlers,
	progressIndicator,

	GOBBLE_INFO    = chalk.cyan( 'gobble:' ) + ' ',
	GOBBLE_ERROR   = chalk.bgRed.white( 'gobble:' ) + ' ',
	GOBBLE_WARNING = chalk.bgYellow.black( 'gobble:' ) + ' ';

function info ( message ) {
	write( GOBBLE_INFO + indent( message, 8 ) );
}

function warn ( message ) {
	write( GOBBLE_WARNING + indent( message, 8 ) );
}

function error ( message ) {
	write( GOBBLE_ERROR + indent( message, 8 ) );
}

function indent ( str, indentAmount ) {
	var numColumns, words, indentStr, lines = [], currentLine;

	if ( !process.stderr.isTTY ) {
		return str;
	}

	numColumns = process.stderr.columns - indentAmount;
	indentStr = new Array( indentAmount + 1 ).join( ' ' );

	words = str.split( ' ' );
	currentLine = words.shift();

	words.forEach( function ( word ) {
		if ( ( plainText( currentLine ).length + plainText( word ).length + 1 ) < numColumns ) {
			currentLine += ' ' + word;
		} else {
			currentLine += '\n';
			lines.push( currentLine );
			currentLine = word;
		}
	});

	lines.push( currentLine );

	return lines.join( indentStr );
}

function plainText ( str ) {
	// strip control codes
	return str.replace( ansiRegex(), '' );
}

function write ( message ) {
	process.stderr.write( message + '\n' );
}

messages = {
	BUILD_INVALIDATED: function ( x ) {
		return f( 'build invalidated (%s). restarting', summariseChanges( x.changes ) );
	},

	BUILD_START: function () {
		return 'build started';
	},

	BUILD_COMPLETE: function ( x ) {
		var result = ( x.dest ? 'built to ' + chalk.bold( x.dest ) : 'build completed' ) + ' in ' + chalk.bold( x.duration + 'ms' );

		if ( x.watch ) {
			result += '. Listening for changes...\n';
		}

		return result;
	},

	GOBBLEFILE_CHANGED: function () {
		return 'gobblefile changed. restarting server';
	},

	MERGE_START: function ( x ) {
		return chalk.bold( x.id ) + ' running...';
	},

	MERGE_COMPLETE: function ( x ) {
		return x.id + ' done in ' + chalk.bold( x.duration + 'ms' );
	},

	LIVERELOAD_RUNNING: function () {
		return 'livereload server running';
	},

	SERVER_LISTENING: function ( x ) {
		return 'server listening on port ' + chalk.bold( x.port );
	},

	TRANSFORM_START: function ( x ) {
		return chalk.bold( x.id ) + ' running...';
	},

	TRANSFORM_COMPLETE: function ( x ) {
		return x.id + ' done in ' + chalk.bold( x.duration + 'ms' );
	}
};

warnings = {
	OLD_GOBBLE: function ( x ) {
		return 'Your locally installed gobble version (' + chalk.bold( x.version ) + ') is out of date, and may cause problems with gobble-cli. You should update with ' + chalk.cyan( 'npm\u00A0update\u00A0gobble' );
	},

	MAP_DEPRECATED: function () {
		return 'node.map() is deprecated. You should use node.transform() instead for both file and directory transforms';
	}
};

errorHandlers = {
	DIR_NOT_EMPTY: function ( err ) {
		error( 'destination directory (' + err.path + ') is not empty! Use ' + chalk.cyan( '--force' ) + ' or ' + chalk.cyan( '-f' ) + ' to continue anyway' );
		return true;
	},

	MISSING_DEST_DIR: function ( err ) {
		error( 'you must specify a destination directory, e.g. ' + chalk.cyan( 'gobble ' + err.task + ' public' ) + ', where \'public\' is the target directory' );
		return true;
	},

	MISSING_DIRECTORY: function ( err ) {
		error( 'the ' + chalk.cyan( err.path ) + ' directory does not exist!' );
	},

	MISSING_GOBBLE: function () {
		error( 'could not find a local copy of gobble. You should probably install it with ' + chalk.cyan( 'npm\u00A0install\u00A0--save-dev\u00A0gobble' ) );
		return true;
	},

	MISSING_GOBBLEFILE: function () {
		error( 'You must have a gobblefile.js in your project\'s root folder in order to use gobble from the command line.\n\nSee ' + chalk.cyan( 'https://github.com/gobblejs/gobble/wiki/How-to-write-a-gobblefile' ) + ' for help getting started' );
		return true;
	},

	PLUGIN_NOT_FOUND: function ( err ) {
		var plugin = err.plugin;

		console.log( GOBBLE_INFO + 'could not load ' + plugin + ' plugin. Have you installed it? ' + chalk.cyan( 'npm\u00A0install\u00A0--save-dev\u00A0gobble-' + plugin ) );

		tryInstallPlugin( plugin, function ( err ) {
			if ( err ) {
				logger.info( 'exiting' );
				process.exit( 1 );
			}

			console.log( GOBBLE_INFO + 'successfully installed ' + plugin + ' plugin' );
			// TODO how to restart on install?
			// restart();
		});
	},

	PORT_IN_USE: function ( err ) {
		error( 'port ' + err.port + ' is already in use. Are you already running gobble? You can specify a different port with e.g. ' + chalk.cyan( 'gobble -p 5678' ) );
		return true; // fatal
	},

	STARTUP_ERROR: function ( err ) {
		error( 'error starting gobble' );

		console.log( chalk.red( '>>>' ) );
		console.log( ( err.original.message || err.original ).trim() );
		console.log( chalk.grey( err.original.stack ) );
		console.log( chalk.red( '<<<' ) );

		return true; // fatal
	},

	TRANSFORMATION_FAILED: function ( err ) {
		error( err.id + ' transformation failed\n' );

		console.log( chalk.red( '>>>' ) );
		console.log( ( err.original.message || err.original ).trim() );
		console.log( chalk.grey( err.stack ) );
		console.log( chalk.red( '<<<' ) );
	}
};

logger = {
	info: function ( details ) {
		var fn, message;

		if ( progressIndicator ) {
			progressIndicator.stop();
			progressIndicator = null;
		}

		if ( fn = messages[ details.code ] ) {
			message = fn( details );
		} else {
			message = util.format.apply( null, arguments );
		}

		if ( details.progressIndicator ) {
			progressIndicator = stevedore({
				message: message,
				frames: frames
			});
		} else {
			info( message );
		}
	},

	error: function ( err ) {
		var errorHandler, fatal;

		if ( progressIndicator ) {
			progressIndicator.stop();
			progressIndicator = null;
		}

		process.stderr.write( '\n' ); // newline

		// allow error to be passed as a string
		if ( typeof err !== 'object' ) {
			error( util.format.apply( util, arguments ) );
			return;
		}

		// Handle known errors
		// TODO err.gobble is deprecated in favour of err.name === 'GobbleError'
		if ( err.name === 'GobbleError' && ( errorHandler = errorHandlers[ err.code ] ) ) {
			fatal = errorHandler( err );
		}

		else {
			error( ( err.message || err ).trim() );

			console.log( chalk.red( '>>>' ) );
			console.log( chalk.grey( err.stack ) );
			console.log( chalk.red( '<<<' ) );

			fatal = true;
		}

		console.log( '' ); // newline

		if ( !fatal ) {
			info( 'listening for changes...\n' );
		}
	},

	warn: function ( details ) {
		var fn, message;

		if ( fn = warnings[ details.code ] ) {
			message = fn( details );
		} else {
			throw new Error( 'gobble-cli received an unexpected warning: ' + util.format.apply( null, arguments ) );
		}

		if ( !alreadyWarned[ message ] ) {
			warn( message );
		}
	}
};

module.exports = logger;
